from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC
from ewah.ewah_utils.airflow_utils import datetime_utcnow_with_tz

from ewah.hooks.base import EWAHBaseHook as BaseHook

from facebook_business.api import FacebookAdsApi
from facebook_business.adobjects.adaccount import AdAccount
from facebook_business.adobjects.adsinsights import AdsInsights

from datetime import datetime, timedelta

import inspect
import time


class EWAHFBOperator(EWAHBaseOperator):

    _NAMES = ["facebook", "fb"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: True,
    }

    class levels:
        ad = "ad"

    def __init__(
        self,
        insight_fields,
        level,
        data_since,
        account_ids=None,
        refresh_interval=timedelta(days=7),
        breakdowns=None,
        *args,
        **kwargs
    ):

        if isinstance(refresh_interval, int):
            refresh_interval = timedelta(days=refresh_interval)
        elif not isinstance(refresh_interval, timedelta):
            raise Exception("refresh_interval must be type timedelta or integer!")

        if kwargs.get("update_on_columns"):
            raise Exception("update_on_columns is set by operator!")

        if (
            not account_ids is None
            and not account_ids.__iter__
            and not isinstance(account_ids, (str, int))
        ):
            raise Exception(
                "account_ids must be a string, integer, or an iterable, such as a list,"
                + " of strings or integers!"
            )
        if account_ids is None or isinstance(account_ids, (str, int)):
            account_ids = [account_ids]

        if level == self.levels.ad:
            kwargs["update_on_columns"] = [
                "ad_id",
                "date_start",
                "date_stop",
            ] + (breakdowns or [])
            insight_fields += ["ad_id", "ad_name"]
            insight_fields = list(set(insight_fields))  # avoid duplications
        else:
            raise Exception("Specified level not supported!")

        allowed_insight_fields = [
            _attr[1]
            for _attr in [
                member
                for member in inspect.getmembers(
                    AdsInsights.Field,
                    lambda a: not (inspect.isroutine(a)),
                )
                if not (member[0].startswith("__") and member[0].endswith("__"))
            ]
        ]
        for i_f in insight_fields:
            if not i_f in allowed_insight_fields:
                raise Exception(
                    (
                        "Field {0} is not an accepted value for insight_fields! "
                        + "Accepted field values:\n\t{1}\n"
                    ).format(i_f, "\n\t".join(allowed_insight_fields))
                )

        assert isinstance(data_since, datetime), "data_from must be of type datetime!"

        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = "date_start"

        super().__init__(*args, **kwargs)

        self.data_since = data_since
        self.account_ids = account_ids
        self.insight_fields = insight_fields
        self.level = level
        self.breakdowns = breakdowns
        self.refresh_interval = refresh_interval

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_since = (
                self.get_max_value_of_column("date_start") - self.refresh_interval
            )
        else:
            data_since = self.data_since
        for account_id in self.account_ids or []:
            for batch in self.source_hook.get_data_in_batches(
                level=self.level,
                fields=self.insight_fields,
                data_from=data_since,
                data_until=datetime.now(),
                account_id=account_id,
                breakdowns=self.breakdowns,
            ):
                self.upload_data(batch)



from airflow.hooks.base import BaseHook
from airflow.models.connection import Connection
from airflow.providers_manager import ProvidersManager
from airflow.utils.module_loading import import_string

from typing import Type, Optional


class EWAHConnection(Connection):
    """Extension of airflow's native Connection."""

    @classmethod
    def get_connection_from_secrets(cls, conn_id):
        """Save the calling hook class as provided when called."""
        conn = super().get_connection_from_secrets(conn_id)
        # conn is returned as airflow.hooks.base.BaseHook, return as cls instead
        conn = cls(
            conn_id=conn.conn_id,
            conn_type=conn.conn_type,
            description=conn.description,
            host=conn.host,
            login=conn.login,
            password=conn.password,
            schema=conn.schema,
            port=conn.port,
            extra=conn.extra,
        )

        # add hook_cls attribute
        hook_class_name = ProvidersManager().hooks.get(
            conn.conn_type, (None, None, None, None)
        )[0]
        if hook_class_name:
            conn.hook_cls = import_string(hook_class_name)
        else:
            conn.hook_cls = EWAHBaseHook
        return conn

    def get_hook(self):
        return self.hook_cls(conn=self)

    def __getattr__(self, name):
        """Enable: use custom widgets like attributes of a connection.

        E.g. have a field api_key -> get it like conn.api_key!
        """
        if not name == "hook_cls" and hasattr(self, "hook_cls"):
            if (
                hasattr(self.hook_cls, "get_connection_form_widgets")
                or hasattr(self.hook_cls, "get_ui_field_behaviour")
                or self.hook_cls._ATTR_RELABEL
            ):
                if name in self.hook_cls._ATTR_RELABEL.keys():
                    return getattr(self, self.hook_cls._ATTR_RELABEL[name])
                widgets = self.hook_cls.get_connection_form_widgets()
                if name in widgets.keys():
                    return self.extra_dejson.get(name)
                long_name = "extra__" + self.conn_type + "__" + name
                if long_name in widgets.keys():
                    return self.extra_dejson.get(long_name)
        if hasattr(super(), "__getattr__"):
            return super().__getattr__(name)
        _msg = "{0} is not an attribute of {1}!"
        _msg += " conn_id: {2}, conn_type: {3}, hook class: {4}"
        raise AttributeError(
            _msg.format(
                name,
                self.__class__.__name__,
                # If accessed directly, attributes would cause infinite loop
                self.__dict__.get("conn_id"),
                self.__dict__.get("conn_type"),
                self.__dict__.get("hook_cls"),
            )
        )


class EWAHBaseHook(BaseHook):
    """Extension of airflow's native Base Hook."""

    # Overwrite in child class to relabel attributes
    # e.g. {"api_key": "password"} to get self.password for self.api_key
    _ATTR_RELABEL = {}

    def __init__(
        self,
        conn: Optional[EWAHConnection] = None,
        conn_id: Optional[str] = None,
        *args,
        **kwargs
    ):
        if not (conn or conn_id):
            raise Exception("Must supply either 'conn' or 'conn_id'!")
        if conn and conn_id:
            raise Exception("Must not supply both 'conn' or 'conn_id'!")
        if conn_id:
            conn = self.get_connection(conn_id)
        self.conn = conn

        if not self.conn_name_attr:
            raise AirflowException("conn_name_attr is not defined")
        elif len(args) == 1:
            setattr(self, self.conn_name_attr, args[0])
        elif self.conn_name_attr not in kwargs:
            setattr(self, self.conn_name_attr, self.default_conn_name)
        else:
            setattr(self, self.conn_name_attr, kwargs[self.conn_name_attr])

        return super().__init__()

    @classmethod
    def get_connection(cls, conn_id: str) -> EWAHConnection:
        """
        Overwrite classmethod to use extended Connection object.

        Get connection, given connection id.
        :param conn_id: connection id
        :return: connection
        """
        return EWAHConnection.get_connection_from_secrets(conn_id)

    @classmethod
    def get_hook_from_conn_id(cls, conn_id: str):
        conn = cls.get_connection(conn_id=conn_id)
        return conn.hook_cls(conn=conn)



from ewah.constants import EWAHConstants as EC
from ewah.operators import operator_list as operators
from ewah.operators.base import EWAHBaseOperator as EBO

from .dag_factory_atomic import dag_factory_atomic
from .dag_factory_idempotent import dag_factory_idempotent
from .dag_factory_mixed import dag_factory_mixed

from airflow.utils.log.logging_mixin import LoggingMixin

from typing import Optional, Any
from copy import deepcopy
from jinja2 import Template

import yaml
from ewah.ewah_utils.yml_loader import Loader, Dumper


class EWAHDAGGenerator(LoggingMixin):
    """
    Generator class to yield all DAGs defined in the dag definition dict.
    """

    DAG_FACTORIES = {
        EC.DS_ATOMIC: dag_factory_atomic,
        "fr": dag_factory_atomic,
        "full-refresh": dag_factory_atomic,
        "full refresh": dag_factory_atomic,
        EC.DS_MIXED: dag_factory_mixed,
        "fullcremental": dag_factory_mixed,
        "periodic fr": dag_factory_mixed,
        EC.DS_IDEMPOTENT: dag_factory_idempotent,
        "idemnpotent": dag_factory_idempotent,
        "inc": dag_factory_idempotent,
        "incr": dag_factory_idempotent,
        "incremental": dag_factory_idempotent,
    }

    @staticmethod
    def _default(my_dict: dict, key: str, default_value: Any):
        "Set default value to a key if key does not exist in dict."
        my_dict[key] = my_dict.pop(key, default_value)

    def __init__(self, dag_definition_dict: dict) -> None:
        self.base_config = dag_definition_dict.pop("base_config", {})
        self.dags_dict = dag_definition_dict.pop("el_dags", {})
        _msg = "Invalid definitions:\n\n{0}".format(str(dag_definition_dict))
        assert not dag_definition_dict, _msg
        assert self.dags_dict, "No DAGs specified in config!"

        self._default(self.base_config, "additional_dag_args", {})
        self._default(self.base_config, "additional_task_args", {})
        self._default(self.base_config["additional_task_args"], "owner", "EWAH")
        self._default(
            self.base_config["additional_task_args"], "weight_rule", "upstream"
        )

    def __iter__(self):
        "Generator to yield individual DAGs as parsed from DAGs dictionary."
        self.log.info("EWAH is now loading DAGs from dict")

        def get_logging_func(dag_name):
            def logging_func(msg):
                self.log.info(
                    "EWAH LOG - DAG: {0} - Message: {1}".format(dag_name, msg)
                )

            return logging_func

        for dag_name, dag_config in deepcopy(self.dags_dict).items():
            self.log.info("EWAH is loading DAG(s): %s", dag_name)

            # Start with base config as default config, build up from there
            config = deepcopy(self.base_config)
            config["logging_func"] = get_logging_func(dag_name)
            config["additional_task_args"].update(
                dag_config.pop("additional_task_args", {})
            )
            config["additional_dag_args"].update(
                dag_config.pop("additional_dag_args", {})
            )
            config.update(dag_config)
            config["dag_name"] = dag_name

            dwh_engine = config.pop("dwh_engine", None)
            if (
                dwh_engine is None
                or not dwh_engine.lower() in EC.DWH_ENGINE_SPELLING_VARIANTS
            ):
                raise Exception(
                    "Invalid dwh_engine {0} for DAG {1}!".format(
                        dwh_engine,
                        dag_name,
                    )
                )
            config["dwh_engine"] = EC.DWH_ENGINE_SPELLING_VARIANTS[dwh_engine]

            el_strategy = config.pop("el_strategy", config.pop("dag_strategy", None))
            if (
                el_strategy is None
                or not el_strategy.lower() in self.DAG_FACTORIES.keys()
            ):
                raise Exception(
                    "Invalid el_strategy {0} for DAG {1}".format(
                        el_strategy,
                        dag_name,
                    )
                )
            dag_factory = self.DAG_FACTORIES[el_strategy.lower()]

            el_operator = config.pop("el_operator", None)
            if isinstance(el_operator, type) and issubclass(el_operator, EBO):
                config["el_operator"] = el_operator
            else:
                if not el_operator.lower() in operators:
                    _msg = "Invalid el_operator {0} for DAG {1}!".format(
                        el_operator,
                        dag_name,
                    )
                    raise Exception(_msg)
                config["el_operator"] = operators[el_operator.lower()]

            dags = dag_factory(**config)
            for dag in dags:
                # Iterate over individual DAGs!
                yield dag


def dags_from_yml_file(file_path: str) -> EWAHDAGGenerator:
    """Load DAGs from a YAML file.

    Returns an initialized generator class that iterates over all the individual
    DAGs as defined in the provided YAML file.

    This YAML file may contain Jinja2 templating.

    :param file_path: Complete path to the YAML file containing the DAGs
        definition.
    """

    return EWAHDAGGenerator(
        dag_definition_dict=yaml.load(open(file_path, "r"), Loader=Loader)
    )



"""
The Snowflake Uploader is currently not available due to questionable design decisions
by the Snowflake Python SDK maintainers.
"""

from ewah.uploaders.base import EWAHBaseUploader
from ewah.constants import EWAHConstants as EC

import os
import csv
import pickle

import snowflake.connector
from tempfile import NamedTemporaryFile, TemporaryDirectory
from airflow.models import BaseOperator


class SnowflakeOperator(BaseOperator):
    "Operate to execute SQL on Snowflake"

    def __init__(self, sql, snowflake_conn_id, database, params=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sql = sql
        self.conn_id = snowflake_conn_id
        self.database = database
        # an attribute called "params" of an operator leads to an airflow error
        # if that attribute is None due to default templating of "task.params"!
        self._params = params

    def execute(self, context):
        from ewah.hooks.snowflake import EWAHSnowflakeHook

        hook = EWAHSnowflakeHook(
            conn_id=self.conn_id,
            database=self.database,
        )
        hook.execute(self.sql, commit=True, params=self._params)
        hook.commit()
        hook.close()


class EWAHSnowflakeUploader(EWAHBaseUploader):

    _QUERY_SCHEMA_CHANGES_COLUMNS = """
        SELECT
            column_name
        FROM "{database_name}".information_schema.columns
        WHERE table_catalog = %(database_name)s
            AND table_schema = %(schema_name)s
            AND table_name = %(table_name)s
        ORDER BY ordinal_position ASC;
    """
    _QUERY_SCHEMA_CHANGES_DROP_COLUMN = """
        ALTER TABLE "{database_name}"."{schema_name}"."{table_name}"
        DROP COLUMN IF EXISTS "{column_name}" CASCADE;
    """
    _QUERY_SCHEMA_CHANGES_ADD_COLUMN = """
        ALTER TABLE "{database_name}"."{schema_name}"."{table_name}"
        ADD COLUMN "{column_name}" {column_type};
    """
    _QUERY_TABLE = """
        SELECT * FROM "{database_name}"."{schema_name}"."{table_name}"
    """

    _COPY_TABLE = """
        CREATE OR REPLACE TABLE "{database_name}"."{new_schema}"."{new_table}"
            CLONE "{database_name}"."{old_schema}"."{old_table}";
    """

    def __init__(self, *args, database=None, **kwargs):
        self.database = database  # TODO: can I remove this??
        super().__init__(EC.DWH_ENGINE_SNOWFLAKE, *args, **kwargs)

    def commit(self):
        self.dwh_hook.commit()

    def rollback(self):
        self.dwh_hook.rollback()

    def close(self):
        self.dwh_hook.close()

    @property
    def tempdir(self):
        if not hasattr(self, "_tempdir"):
            self._tempdir = TemporaryDirectory(prefix="uploadtosnowflake")
        return self._tempdir.name

    def _create_or_update_table(
        self,
        data,
        table_name,
        schema_name,
        schema_suffix,
        columns_definition,
        columns_partial_query,
        update_on_columns,
        load_strategy,
        upload_call_count,
        database_name=None,
        pk_columns=None,
    ):
        database_name = database_name or self.dwh_hook.conn.database
        pk_columns = pk_columns or []
        self.log.info("Preparing DWH Tables...")
        schema_name += schema_suffix
        new_table_name = table_name + "_new"
        self.dwh_hook.execute(
            sql="""CREATE OR REPLACE TABLE
                    "{database_name}"."{schema_name}"."{table_name}"
                    ({columns});
            """.format(
                database_name=database_name,
                schema_name=schema_name,
                table_name=new_table_name,
                columns=columns_partial_query,
            ),
            commit=False,
        )

        list_of_columns = self.dwh_hook.execute_and_return_result(
            sql="""SELECT
                column_name
                FROM "{0}".information_schema.columns
                WHERE table_catalog = '{0}'
                AND table_schema = '{1}'
                AND table_name = '{2}'
                ORDER BY ordinal_position ASC
            """.format(
                database_name,
                schema_name,
                new_table_name,
            ),
        )
        list_of_columns = [col[0].strip() for col in list_of_columns]

        self.log.info("Writing data to a temporary .csv file")
        with TemporaryDirectory(prefix="uploadtosnowflake") as tmp_dir:
            with NamedTemporaryFile(
                dir=tmp_dir,
                prefix=new_table_name,
                suffix=".csv",
            ) as datafile:
                file_name = os.path.abspath(datafile.name)
                with open(file_name, mode="w") as csv_file:
                    csvwriter = csv.writer(
                        csv_file,
                        delimiter=",",
                        quotechar='"',
                        quoting=csv.QUOTE_MINIMAL,
                    )

                    for _ in range(len(data)):
                        datum = data.pop(0)
                        # Make sure order of csv is the same as order of columns
                        csvwriter.writerow(
                            [datum.get(col) for col in list_of_columns],
                        )

                # now stage and copy into snowflake!
                sql_upload = """
                    USE DATABASE "{2}";
                    USE SCHEMA "{3}";
                    DROP FILE FORMAT IF EXISTS {1}_format;
                    CREATE FILE FORMAT {1}_format
                        type = 'CSV'
                        field_delimiter = ','
                        field_optionally_enclosed_by = '"'
                    ;
                    DROP STAGE IF EXISTS {1}_stage;
                    CREATE STAGE {1}_stage
                        file_format = {1}_format;
                    PUT file://{0} @{1}_stage AUTO_COMPRESS = TRUE OVERWRITE = TRUE;
                    COPY INTO "{2}"."{3}"."{1}" FROM '@{1}_stage/{4}.gz';
                """.format(
                    file_name,
                    new_table_name,
                    database_name,
                    schema_name,
                    file_name[file_name.rfind("/") + 1 :],
                )
                self.log.info("Uploading data to Snowflake...")
                self.dwh_hook.execute(sql_upload)

        if (load_strategy == EC.LS_INSERT_REPLACE and upload_call_count == 1) or (
            not self.test_if_table_exists(
                table_name=table_name,
                schema_name=schema_name,
                database_name=database_name,
            )
        ):
            sql_final = """
                USE SCHEMA "{0}"."{1}";
                DROP TABLE IF EXISTS "{0}"."{1}"."{2}" CASCADE;
                ALTER TABLE "{0}"."{1}"."{3}" RENAME TO "{2}";
            """.format(
                database_name,
                schema_name,
                table_name,
                new_table_name,
            )
            if pk_columns:
                sql_final += """
                    ALTER TABLE "{0}"."{1}"."{2}"
                    ADD PRIMARY KEY ("{3}");
                """.format(
                    database_name,
                    schema_name,
                    table_name,
                    '","'.join(pk_columns),
                )
        else:
            update_set_cols = []
            for col in columns_definition.keys():
                if not (col in update_on_columns):
                    update_set_cols += [col]

            sql_final = """
                USE SCHEMA "{0}"."{1}";
                MERGE INTO "{0}"."{1}"."{2}" AS a
                    USING "{0}"."{1}"."{3}" AS b
                    ON {4}
                    WHEN MATCHED THEN UPDATE
                        SET {5}
                    WHEN NOT MATCHED THEN INSERT
                                ({6})
                        VALUES  ({7})
                    ;
                DROP TABLE "{0}"."{1}"."{3}" CASCADE;
            """.format(
                database_name,
                schema_name,
                table_name,
                new_table_name,
                " AND ".join(
                    ['a."{0}" = b."{0}"'.format(col) for col in update_on_columns]
                )
                or "FALSE",
                ", ".join(['a."{0}" = b."{0}"'.format(col) for col in update_set_cols]),
                '"' + '", "'.join(list(columns_definition.keys())) + '"',
                ", ".join(
                    ['b."{0}"'.format(col) for col in list(columns_definition.keys())]
                ),
            )

        self.log.info("Final Step: Merging data")
        self.dwh_hook.execute(sql_final)

    def test_if_table_exists(
        self,
        table_name,
        schema_name,
        database_name=None,
    ):
        self.dwh_hook.execute(
            "USE DATABASE {0}".format(
                database_name or self.database or self.dwh_hook.conn.database,
            )
        )
        return 0 < len(
            self.dwh_hook.execute_and_return_result(
                """
            SELECT * FROM "{0}".information_schema.tables
            WHERE table_schema LIKE '{1}'
            AND table_name LIKE '{2}'
            """.format(
                    database_name or self.database or self.dwh_hook.conn.database,
                    schema_name,
                    table_name,
                )
            )
        )

    def get_max_value_of_column(
        self, column_name, table_name, schema_name, database_name=None
    ):
        self.dwh_hook.execute(
            "USE DATABASE {0}".format(
                database_name or self.database or self.dwh_hook.conn.database,
            )
        )
        return self.dwh_hook.execute_and_return_result(
            sql='SELECT MAX("{0}") FROM {1}'.format(
                column_name,
                f'"{schema_name}"."{table_name}"',
            ),
            return_dict=False,
        )[0][0]



from ewah.hooks.base import EWAHBaseHook

from collections import OrderedDict
from simple_salesforce import Salesforce, format_soql
from typing import Union, List, Dict, Optional, Any
from datetime import datetime

import requests


class EWAHSalesforceHook(EWAHBaseHook):

    _ATTR_RELABEL: dict = {
        "username": "login",
        "domain": "schema",
    }

    conn_name_attr: str = "ewah_salesforce_conn_id"
    default_conn_name: str = "ewah_salesforce_default"
    conn_type: str = "ewah_salesforce"
    hook_name: str = "EWAH Salesforce Connection"

    _FIELDS_CACHE = {}

    _OAUTH_URL = "https://{0}.salesforce.com/services/oauth2/token"

    version = "49.0"

    @property
    def sf_conn(self) -> Salesforce:
        """Return an initialized Salesforce object based on the airflow connection."""

        if not hasattr(self, "_sf_conn"):
            self.log.info(
                f"Initializing connection with username {self.conn.username}!"
            )
            if self.conn.client_secret:
                response = requests.post(
                    self._OAUTH_URL.format(self.conn.domain or "login"),
                    data={
                        "client_secret": self.conn.client_secret,
                        "client_id": self.conn.client_id,
                        "grant_type": "password",
                        "username": self.conn.username,
                        "password": self.conn.password
                        + (self.conn.security_token or ""),
                    },
                ).json()
                self._sf_conn = Salesforce(
                    instance_url=response["instance_url"],
                    session_id=response["access_token"],
                    version=self.version,
                )
            else:
                self._sf_conn = Salesforce(
                    username=self.conn.username,
                    password=self.conn.password,
                    security_token=self.conn.security_token,
                    domain=self.conn.domain or None,
                    client_id=self.conn.client_id or None,
                    version=self.version,
                )
        return self._sf_conn

    def get_incrementer(self, salesforce_object):
        # Figure out what field to use for date comparison
        fields = self.get_fields_of_object(salesforce_object)
        options = ["SystemModstamp", "LastModifiedDate", "CreatedDate"]
        for option in options:
            if option in fields:
                return option

    def get_data_in_batches(
        self,
        salesforce_object: str,
        columns: Optional[list] = None,
        data_from: Optional[datetime] = None,
        data_until: Optional[datetime] = None,
        batch_size: int = 10000,
    ) -> List[Dict[str, Any]]:
        """Generator to return all data of a salesforce object"""

        # Figure out what field(s) to use for date comparison & data loading
        fields = self.get_fields_of_object(salesforce_object)
        incrementer = self.get_incrementer(salesforce_object=salesforce_object)

        # Create SOQL query
        query = "SELECT\n\t{0}\nFROM {1}".format(
            ",\n\t".join(columns or fields), salesforce_object
        )
        where_clauses = []
        if data_from and incrementer:
            where_clauses.append(f"{incrementer} >= " + format_soql("{0}", data_from))
        if data_until and incrementer:
            where_clauses.append(f"{incrementer} < " + format_soql("{0}", data_until))
        if where_clauses:
            query += "\nWHERE " + "\n  AND ".join(where_clauses)
        self.log.info(f"Querying data with this SOQL query:\n\n{query}")

        # Yield results
        result = self.sf_conn.query(query, include_deleted=True)
        data = []
        while (result.get("done") and result.get("records")) or result.get(
            "nextRecordsUrl"
        ):
            result_data = result.pop("records")
            for datum in result_data:
                del datum["attributes"]
                datum = dict(datum)
            data += result_data
            next_page = result.get("nextRecordsUrl")
            if next_page:
                result = self.sf_conn.query_more(next_page, True)
            if len(data) >= batch_size:
                yield data
                data = []

        if data:
            yield data

        # Make sure there was no error on the way
        if not result.get("done"):
            raise Exception("SOQL query error! Response: {0}".format(result))

    def get_soql_query_result(
        self, soql_query: str, remove_attribues: bool = True, convert_dict: bool = False
    ) -> Union[List[OrderedDict], List[dict]]:
        """Returns the result of an SOQL query.

        Avoid running multiple large queries simultaneously. Queries are paginated and
        Salesforce will only ever make the last 10 pages available. If running too many
        queries at once, this may mean that a page will become unavailable before it
        was fetched, failing the task.

        :param soql_query: SOQL query to run and return result.
        :param remove_attribues: If TRUE, removes "attributes" field from data
            dictionary.
        :param convert_dict: If TRUE, converts result in a list of normal dictionaries.
            Otherwise, will return a list of OrderedDict.

        """
        result = self.sf_conn.query_all(soql_query)

        if not result.get("done"):
            raise Exception("SOQL query error! Response: {0}".format(result))

        data = result["records"]

        if data and (remove_attribues or convert_dict):
            for datum in data:
                if remove_attribues:
                    del datum["attributes"]
                if convert_dict:
                    datum = dict(datum)

        return data

    def get_fields_of_object(
        self, salesforce_object: str, return_types: bool = False
    ) -> Union[List[str], Dict[str, str]]:
        """Returns a list or dict of all fields of a Salesforce object.

        :param salesforce_object: Name of the Salesforce object (e.g. Account)
        :param return_types: If TRUE, return a dict of field_name:field_type; if FALSE,
            return a list of field names.
        """
        if self._FIELDS_CACHE.get(salesforce_object):
            fields = self._FIELDS_CACHE[salesforce_object]
        else:
            fields = {
                field["name"]: field["type"]
                for field in getattr(self.sf_conn, salesforce_object).describe()[
                    "fields"
                ]
            }
            self._FIELDS_CACHE[salesforce_object] = fields

        if return_types:
            return fields
        else:
            return [key for (key, value) in fields.items()]

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["port", "extra", "host"],
            "relabeling": {
                "login": "Username",
                "password": "Password",
                "schema": "Domain (default: 'login')",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from wtforms import StringField

        return {
            "extra__ewah_salesforce__security_token": StringField(
                "Security Token",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_salesforce__client_id": StringField(
                "Client ID (optional)",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_salesforce__client_secret": StringField(
                "Client Secret (optional)",
                widget=BS3TextFieldWidget(),
            ),
        }



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

import gspread
import json
from oauth2client.service_account import ServiceAccountCredentials as SAC


class EWAHGSpreadOperator(EWAHBaseOperator):

    _NAMES = ["google_sheets", "gs", "gsheets"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _REQUIRES_COLUMNS_DEFINITION = True

    _SAMPLE_JSON = {
        "client_secrets": {
            "type": "service_account",
            "project_id": "abc-123",
            "private_key_id": "123456abcder",
            "private_key": "-----BEGIN PRIVATE KEY-----\nxxx\n-----END PRIVATE KEY-----\n",
            "client_email": "xyz@abc-123.iam.gserviceaccount.com",
            "client_id": "123457",
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/xyz%40abc-123.iam.gserviceaccount.com",
        }
    }

    def _translate_alphanumeric_column(self, column_identifier):
        if type(column_identifier) == str:
            column_number = 0
            i = 0
            ident_dict = {}
            while column_identifier:
                letter = column_identifier[-1:].lower()
                if ord(letter) > ord("z") or ord(letter) < ord("a"):
                    raise Exception(
                        "Column letter {0} out of bounds!".format(
                            letter,
                        )
                    )
                column_identifier = column_identifier[:-1]
                ident_dict.update({i: ord(letter) + 1 - ord("a")})
                i += 1
            return sum([v * (26 ** k) for k, v in ident_dict.items()])
        else:
            return column_identifier

    def __init__(
        self,
        workbook_key,  # can be seen in the URL of the workbook
        sheet_key,  # name of the worksheet
        start_row=2,  # in what row does the data begin?
        end_row=None,  # optional: what is the last row? None gets all data
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        credentials = BaseHook.get_connection(self.source_conn_id).extra_dejson
        credentials = credentials.get("client_secrets", credentials)

        _msg = "Google Service Account Credentials misspecified!"
        _msg += " Example of a correct specifidation: {0}".format(
            json.dumps(self._SAMPLE_JSON)
        )
        for key in self._SAMPLE_JSON["client_secrets"]:
            if not key in credentials:
                raise Exception(_msg)

        column_match = {}
        for col_key, col_def in self.columns_definition.items():
            if (not col_def) or (not col_def.get(EC.QBC_FIELD_GSHEET_COLNO)):
                raise Exception(
                    (
                        "Column {0} is missing information regarding the "
                        + "position of the column in the sheet."
                    ).format(col_key)
                )
            column_match.update(
                {
                    self._translate_alphanumeric_column(
                        col_def[EC.QBC_FIELD_GSHEET_COLNO],
                    ): col_key,
                }
            )

        self.client_secrets = credentials
        self.column_match = column_match
        self.workbook_key = workbook_key
        self.sheet_key = sheet_key
        self.start_row = start_row
        self.end_row = end_row

    def ewah_execute(self, context):
        client = gspread.authorize(
            SAC.from_json_keyfile_dict(
                self.client_secrets,
                ["https://spreadsheets.google.com/feeds"],
            ),
        )

        self.log.info("Retrieving data...")
        workbook = client.open_by_key(self.workbook_key)
        sheet = workbook.worksheet(self.sheet_key)
        raw_data = sheet.get_all_values()[self.start_row - 1 : self.end_row]

        # Load the data from the sheet into a format for upload into the DWH
        data = []
        for row in raw_data:  # Iterate through each row
            data_dict = {}  # New row, new dictionary of field:value
            row_is_null = True  # Is entire row empty?
            for position, column in self.column_match.items():
                datapoint = row[position - 1]
                data_dict.update({column: datapoint})
                if row_is_null:
                    if bool(datapoint):
                        if not (datapoint == "0"):
                            # Special case with zeroes in text format
                            row_is_null = False  # Row is not empty!
            if not row_is_null:  # Ignore empty rows
                data.append(data_dict)

        self.upload_data(data)



from ewah.hooks import hook_class_names, connection_types


def get_provider_info():
    return {
        "package-name": "ewah",
        "name": "ewah",
        "description": "",
        "hook-class-names": hook_class_names,
        # hook-class-names is deprecated from 2.2.0 and will be removed in airflow 3.0
        # use connection-types instead, keep hook-class-names for backward compatibility
        "connection_types": connection_types,
        "versions": ["1.0.0"],
    }



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.mailchimp import EWAHMailchimpHook


class EWAHMailchimpOperator(EWAHBaseOperator):

    _NAMES = ["mailchimp", "mc"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _REQUIRES_COLUMNS_DEFINITION = False

    _CONN_TYPE = EWAHMailchimpHook.conn_type

    def __init__(self, resource=None, *args, **kwargs):
        # use target table name as resource if none is given (-> easier config)
        self.resource = resource or kwargs.get("target_table_name")
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context):
        for batch in self.source_hook.get_data_in_batches(resource=self.resource):
            self.upload_data(batch)
